let blk = try int_of_string Sys.argv.(1) with _ -> 64
let res = try int_of_string Sys.argv.(2) with _ -> 16
let nth = try int_of_string Sys.argv.(3) with _ -> 8
let mxi = try int_of_string Sys.argv.(4) with _ -> 512

let width = blk * res and height = blk * res

let wx1 = ref (try float_of_string Sys.argv.(5) with _ -> -2.0)
and wy1 = ref (try float_of_string Sys.argv.(6) with _ -> -1.5)
and wx2 = ref (try float_of_string Sys.argv.(7) with _ -> 1.0)
and wy2 = ref (try float_of_string Sys.argv.(8) with _ -> 1.5)

let mandel_point x0 y0 =
  let x0 = (float x0 /. float width) *. (!wx2 -. !wx1) +. !wx1 in
  let y0 = (float y0 /. float height) *. (!wy2 -. !wy1) +. !wy1 in
  let rec mandel n x y =
    if n > mxi then
      mxi
    else
      let x' = x *. x -. y *. y +. x0 in
      let y' = 2. *. x *. y +. y0 in
	if x' *. x' +. y' *. y' > 4. then
	  n
	else
	  mandel (n + 1) x' y'
  in mandel 0 0. 0.

let mandel_fragment x1 y1 x2 y2 =
  let img =
    Array.make_matrix (y2 - y1) (x2 - x1)
      (12)
  in
    for y = y1 to y2 - 1 do
      for x = x1 to x2 - 1 do
	let v = (mandel_point x y * mxi) / 255 in
	  img.(y2 - y - 1).(x - x1) <-  v 
      done
    done

let queue, launch =
  let m = Mutex.create () in
  let q = ref [] in
  let pop () =
    Mutex.lock m ;
    match !q with
      | t :: ts -> q := ts ; Mutex.unlock m ; Some t
      | [] -> Mutex.unlock m ; None
  in
    ((fun v ->
	Mutex.lock m ;
	q := v :: !q ;
	Mutex.unlock m),
     (fun n ->
	let rec th id =
	  match pop () with
	    | Some t ->
		(* Printf.fprintf stderr
		     "Thread %d picked up a task\n%!" id ; *)
		(try
		   t ()
		 with e ->
		   Printf.fprintf stderr
		     "Task aborted on thread %d: %s\n%!"
		     id (Printexc.to_string e));
		th id
	    | None ->
		(* Printf.fprintf stderr "Thread %d exited\n%!" id ; *)
		()
	in
	let rec launch n =
	  if n <> 0 then (
	    let t = Thread.create th n in
	      launch (n - 1) ;
	      Thread.join t
	  )
	in
	  launch n))

let rec loop () =
  for x = 0 to res - 1 do
    for y = 0 to res - 1 do
      queue (fun () ->
	       mandel_fragment
		 (blk * x) (blk * y)
		 (blk * (x + 1)) (blk * (y + 1)))
    done
  done ;
  launch nth
    
let _ = loop ()


